/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000, 2001  Pan Development Team <pan@rebelbase.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <config.h>
#include <errno.h>

#include <gnome.h>

#include <pan/base/argset.h>
#include <pan/base/debug.h>
#include <pan/base/file-grouplist.h>
#include <pan/base/fnmatch.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/serverlist.h>

#include <pan/articlelist.h>
#include <pan/dialogs/dialogs.h>
#include <pan/globals.h>
#include <pan/grouplist.h>
#include <pan/group-ui.h>
#include <pan/gui.h>
#include <pan/prefs.h>
#include <pan/queue.h>
#include <pan/task-grouplist.h>
#include <pan/util.h>

#include "xpm/folder.xpm"

static gint grouplist_update_ui_mainthread (gpointer);
static void group_menu_popup (GdkEventButton * event);
static void grouplist_filter_changed_cb_nolock (void);
static void grouplist_refresh_filter_nolock (void);
static void grouplist_update_groups (const Group ** groups, int group_qty);
static void grouplist_set_server_nolock (Server *);

extern GtkWidget   * ttips;
static GnomePixmap * folder_pixmap           = NULL;
static GnomePixmap * subscribed_pixmap       = NULL;

static int           group_mode              = GROUP_ALL;
static GtkWidget   * group_clist_menu        = NULL;
static Server      * my_server               = NULL;
static GtkWidget   * grouplist_mode_menu     = NULL;
static GtkWidget   * groupname_filter_entry  = NULL;
static gchar       * groupname_filter_string = NULL;

/**
 * This lock must be gotten _before_ the gui lock, and released
 * _after_ the gui lock is released.  Otherwise deadlock may occur.
 */
static GStaticMutex  grouplist_ui_mutex      = G_STATIC_MUTEX_INIT;

PanCallback * grouplist_server_set_callback = NULL;
PanCallback * grouplist_group_selection_changed = NULL;

static void
grouplist_get_group_readable_name (const Group * group, gchar * buf, gint buflen)
{
	const gchar * readable = group_get_readable_name (group);
	gint bufleft;

	if (!collapse_group_names)
		strncpy (buf, readable, buflen);
	else
	{
		gint pos = 0;
		const gchar * pch = NULL;
		const gchar * march = readable;
		const gchar * last = "";
		gint len = 0;

		while (pos+2<buflen && get_next_token_run (march, '.', &march, &pch, &len)) {
			last = pch;
			if (len > 0) {
				buf[pos++] = *pch;
				buf[pos++] = '.';
			}
		}
		if (pos != 0)
			pos -= 2;
		buf[pos] = '\0';

		if (is_nonempty_string(last))
			strncpy (buf+pos, last, buflen-pos);
	}
	buf[buflen-1] = '\0';
	bufleft = buflen - strlen(buf);

	if (group_is_moderated (group)) {
		strncat (buf, _(" (Moderated)"), bufleft);
		buf[buflen-1] = '\0';
		bufleft = buflen - strlen(buf);
	}

	if (group_is_read_only (group)) {
		strncat (buf, _(" (Read-Only)"), bufleft);
		buf[buflen-1] = '\0';
		bufleft = buflen - strlen(buf);
	}
}

static Server *
grouplist_get_visible_server (void)
{
	return group_mode==GROUP_FOLDERS
		? serverlist_get_named_server(INTERNAL_SERVER_NAME)
		: my_server;
}

static void
group_dialog_clicked_cb (GnomeDialog * dialog, int num, Server * server)
{
	if (num==0) /* yes, get list */
		queue_add (TASK(task_grouplist_new (server, GROUPLIST_ALL)));
}


/*****
******
*****/

/*
 * This thing is frustrating... basically we're trying to reconcile
 * the desired behavior with GtkCList's quirks.  The desired behavior:
 *
 * (1) If the user selects a range, don't activate the group
 *     because it's more than likely that the user is selecting
 *     the range to perform a menu popup operation.
 *
 * (2) If a group is single-clicked and single-click prefs are on,
 *     or if a group was double-clicked, activate it.
 *
 * The problem with (1) is that the selection callback fires once per
 * selected item, so there's no way to tell if the first callback is
 * part of a range being selected or not.  So the workaround here is
 * to push the work off to an idle function that gets invoked after
 * all the selection callbacks are done, then query the grouplist for
 * selection info and fire the grouplist selection pan_callback.
 * (2) isn't as hard to do -- the * variable `button_click_count' is
 * used to keep information from the * keypress callback so that we can
 * use it in the selection callback.
 */
static gint button_click_count = -1;
static gint mb = -1;

static void
grouplist_selection_changed_cb (gpointer call_object,
                                gpointer call_arg,
                                gpointer user_data)
{
	GPtrArray * groups = (GPtrArray*)call_arg;
	Group * single_group = NULL;
	debug_enter ("grouplist_selection_changed_cb");

	/* load the specified group */
	if (groups!=NULL && groups->len==1)
		single_group = GROUP(g_ptr_array_index(groups,0));

	if ((single_group!=NULL)
		&& ((button_click_count==2) /* double click */
			|| (button_click_count==1 /*&& Pan.viewmode==GUI_PANED*/
					&& (  (mb==2 && grouplist_mb2_action==CLICK_ACTION_LOAD)
					    ||(mb==1 && grouplist_mb1_action==CLICK_ACTION_LOAD)))))
	{
		gui_page_set (ARTICLELIST_PAGE);
		pan_lock ();
		articlelist_set_group_nolock (single_group);
		pan_unlock ();
	}

	/* reset the click count for next time... */
	button_click_count = -1;
	mb = -1;

	debug_exit ("grouplist_selection_changed_cb");
}

static void
grouplist_button_press (GtkWidget *widget, GdkEventButton * event)
{
	debug_enter ("grouplist_button_press");
	switch (event->button)
	{
		case 1:
		case 2:
			mb = event->button;
			if (event->type == GDK_2BUTTON_PRESS)
				button_click_count = 2;
			else
				button_click_count = 1;
			break;
		case 3:
			group_menu_popup (event);
			break;
	}
	debug_exit ("grouplist_button_press");
}

static void
grouplist_key_press (GtkWidget     * widget,
                     GdkEventKey   * event,
                     gpointer        user_data)
{
	debug_enter ("grouplist_key_press");
	switch (event->keyval)
	{
		case GDK_Up: case GDK_Down:
		case GDK_Left: case GDK_Right:
		case GDK_Page_Up: case GDK_Page_Down:
			button_click_count = 1;
			break;
		default:
			button_click_count = -1;
	} 
	debug_exit ("grouplist_key_press");
}

static gboolean select_callback_pending = FALSE;

static gint
select_row_cb_mainthread (gpointer data)
{
	GPtrArray * groups;
	debug_enter ("select_row_cb_mainthread");

       	groups = grouplist_get_selected_groups ();
	pan_callback_call (grouplist_group_selection_changed,
	                   Pan.group_clist,
			   groups);
	select_callback_pending = FALSE;
	g_ptr_array_free (groups, TRUE);

	debug_exit ("select_row_cb_mainthread");
	return FALSE;
}
static void
select_row_cb (GtkWidget  * widget,
               int          row,
               int          col,
               GdkEvent   * event)
{
	debug_enter ("select_row_cb");

	if (!select_callback_pending)
	{
		select_callback_pending = TRUE;
		run_in_main_thread_nolock (select_row_cb_mainthread, NULL);
	}

	debug_exit ("select_row_cb");
}

/*****
******
*****/

static void
grouplist_set_selected_groups_nolock (const GPtrArray * a)
{
	GtkCList * clist;
	debug_enter ("grouplist_set_selected_groups_nolock");

	if (a != NULL)
	{
		guint i;

		clist = GTK_CLIST(Pan.group_clist);
		gtk_clist_freeze (clist);
		gtk_clist_unselect_all (clist);
		for (i=0; i!=a->len; ++i)
		{
			const Group * g = GROUP(g_ptr_array_index(a,i));
			const int row = gtk_clist_find_row_from_data (clist, (gpointer)g);
			gtk_clist_select_row (clist, row, 0);
			if (!gtk_clist_row_is_visible (clist, row))
			{
				gfloat valign = 0.5;
				gfloat halign = 0.0;
				gtk_clist_moveto (clist, row, 0, (double)valign, (double)halign);
			}
		}
		gtk_clist_thaw (clist);
	}

	debug_exit ("grouplist_set_selected_groups_nolock");
}

Group*
grouplist_get_selected_group (void)
{
	GList * l;
	GtkCList * clist;
	Group * retval = NULL;
	debug_enter ("grouplist_get_selected_group");

	pan_lock ();
	clist = GTK_CLIST(Pan.group_clist);
	l = clist->selection;
	if (l != NULL)
		retval = GROUP(gtk_clist_get_row_data(clist, GPOINTER_TO_INT(l->data)));
	pan_unlock ();

	debug_exit ("grouplist_get_selected_group");
	return retval;
}

GPtrArray*
grouplist_get_selected_groups (void)
{
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GPtrArray * retval = g_ptr_array_new ();
	GList * l = NULL;
	debug_enter ("grouplist_get_selected_groups");

	/* get the group selected items; otherwise, 
	   use the articlelist's group */
	pan_lock ();
	for (l=clist->selection; l!=NULL; l=l->next)
		g_ptr_array_add (retval, gtk_clist_get_row_data (clist, GPOINTER_TO_INT (l->data)));
	pan_unlock ();

	if (!retval->len) {
		Group * group = articlelist_get_group ();
		if (group != NULL)
			g_ptr_array_add (retval, group);
	}

	debug_exit ("grouplist_get_selected_groups");
	return retval;
}

/*****
******
*****/

static void
grouplist_remove_row (const Group *group)
{
	debug_enter ("grouplist_remove_row");

	if (group->server == grouplist_get_visible_server())
	{
		GtkCList *list = GTK_CLIST(Pan.group_clist);
		int row;

		g_static_mutex_lock (&grouplist_ui_mutex);
		pan_lock();
		row = gtk_clist_find_row_from_data (list, (gpointer)group);
		if (row != -1)
			gtk_clist_remove (list, row);
		pan_unlock();
		g_static_mutex_unlock (&grouplist_ui_mutex);
	}

	debug_exit ("grouplist_remove_row");
}

/*****
******
*****/


void
grouplist_get_all (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (TASK(task_grouplist_new (my_server, GROUPLIST_ALL)));
}

void
grouplist_get_new (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (TASK(task_grouplist_new (my_server, GROUPLIST_NEW)));
}

/*****
******
*****/



void
grouplist_select_all (void)
{
	pan_lock();
	gtk_clist_select_all (GTK_CLIST (Pan.group_clist));
	pan_unlock();
}

void
grouplist_deselect_all (void)
{
	pan_lock ();
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));
	pan_unlock ();
}

/*****
******
*****/

static void
grouplist_set_view_mode_nolock_impl (gint view_mode)
{
	GtkWidget *menu = NULL;
	debug_enter ("grouplist_set_view_mode_nolock_impl");

        menu = gtk_option_menu_get_menu (GTK_OPTION_MENU(grouplist_mode_menu));
        gtk_object_ref (GTK_OBJECT(menu));
        gtk_option_menu_remove_menu (GTK_OPTION_MENU (grouplist_mode_menu));
	switch (group_mode = view_mode) {
		case GROUP_ALL:
			gtk_menu_set_active (GTK_MENU (menu), 0);
			break;
		case GROUP_SUBSCRIBED:
			gtk_menu_set_active (GTK_MENU (menu), 1);
			break;
		case GROUP_NEW:
			gtk_menu_set_active (GTK_MENU (menu), 2);
			break;
		case GROUP_FOLDERS:
			gtk_menu_set_active (GTK_MENU (menu), 3);
			break;
		default: break;
	}
        gtk_option_menu_set_menu (GTK_OPTION_MENU(grouplist_mode_menu), menu);
        gtk_object_unref (GTK_OBJECT(menu));

	debug_exit ("grouplist_set_view_mode_nolock_impl");
}

static void
grouplist_set_view_mode_nolock (GtkWidget *widget, gpointer data)
{
	grouplist_set_view_mode_nolock_impl (GPOINTER_TO_INT(data));
	grouplist_refresh_nolock ();
}

typedef struct
{
	gchar * name;
	int mode;
}
GrouplistModeMenuStruct;

static GtkWidget*
grouplist_mode_menu_create (void)
{
	GtkWidget *option_menu = gtk_option_menu_new();
	GtkWidget *menu = gtk_menu_new ();
	int index = 0;
	int i;
	GrouplistModeMenuStruct foo[] = {
		{NULL, GROUP_ALL},
		{NULL, GROUP_SUBSCRIBED},
		{NULL, GROUP_NEW},
		{NULL, GROUP_FOLDERS}
	};
	const int row_qty = sizeof(foo) / sizeof(foo[0]);
	foo[0].name = _("All Groups");
	foo[1].name = _("Subscribed");
	foo[2].name = _("New Groups");
	foo[3].name = _("Folders");

	for (i=0; i<row_qty; ++i) {
		GtkWidget *item = gtk_menu_item_new_with_label (foo[i].name);
		gtk_widget_show (item);
		gtk_menu_append (GTK_MENU (menu), item);
		gtk_signal_connect (GTK_OBJECT(item), "activate",
				    (GtkSignalFunc)grouplist_set_view_mode_nolock,
				    GINT_TO_POINTER(foo[i].mode));
		if (group_mode == foo[i].mode)
			index = i;
	}

	gtk_menu_set_active (GTK_MENU(menu), index);
        gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
	gtk_widget_show_all (GTK_WIDGET(option_menu));

	return option_menu;
}

/*****
******  CALLBACKS
*****/

static void
grouplist_server_groups_added_cb (gpointer server, gpointer groups, gpointer foo)
{
	Server * visible;
	debug_enter ("grouplist_server_groups_added_cb");

	visible = grouplist_get_visible_server ();
	if (visible!=NULL && visible==SERVER(server))
		grouplist_refresh_nolock ();

	debug_exit ("grouplist_server_groups_added_cb");
}

static void
grouplist_server_groups_removed_cb (gpointer server, gpointer groups, gpointer foo)
{
	Server * visible;
	debug_enter ("grouplist_server_groups_removed_cb");

	visible = grouplist_get_visible_server ();
	if (visible!=NULL && visible==SERVER(server)) {
		GPtrArray * a = (GPtrArray*) groups;
		pan_g_ptr_array_foreach (a, (GFunc)grouplist_remove_row, NULL);
	}

	debug_exit ("grouplist_server_groups_removed_cb");
}

static void
grouplist_groups_changed_cb (gpointer gp_groups, gpointer gp_qty, gpointer unused2)
{
	gint qty = GPOINTER_TO_INT(gp_qty);
	const Group ** groups = (const Group**) gp_groups;

	g_return_if_fail (qty>0);
	g_return_if_fail (groups!=NULL);

	if (groups[0]->server == grouplist_get_visible_server())
		grouplist_update_groups (groups, qty);
}

static void
grouplist_server_activated_cb (gpointer server_data, gpointer unused1, gpointer unused2)
{
	pan_lock ();
	grouplist_set_server_nolock (SERVER(server_data));
	pan_unlock ();
}

/*****
******
*****/

void
grouplist_shutdown_module (void)
{
	/* stop listening... */
	pan_callback_remove (serverlist_get_server_activated_callback(),
	                     grouplist_server_activated_cb, NULL);
	pan_callback_remove (group_get_groups_changed_callback(),
	                     grouplist_groups_changed_cb, NULL);
	pan_callback_remove (grouplist_group_selection_changed,
	                     grouplist_selection_changed_cb, NULL);
	pan_callback_remove (server_get_groups_added_callback(),
	                     grouplist_server_groups_added_cb, NULL);
	pan_callback_remove (server_get_groups_removed_callback(),
	                     grouplist_server_groups_removed_cb, NULL);

	/* remember server so that we can default to it next time */
	if (my_server!=NULL && is_nonempty_string(my_server->name))
		gnome_config_set_string ("/Pan/State/Server", my_server->name);

	gnome_config_sync ();
}

gpointer
grouplist_create (void)
{
	GtkWidget * w;
        GtkWidget * vbox;
        GtkWidget * hbox;
	GtkWidget * toolbar;
	GtkCList * clist;
	char * titles[5];
	debug_enter ("grouplist_create");

	titles[0] = "";
	titles[1] = _("Groups");
	titles[2] = _("Unread");
	titles[3] = _("Total");
	titles[4] = _("Description");

	/* create callbacks */
	grouplist_group_selection_changed = pan_callback_new ();
	grouplist_server_set_callback = pan_callback_new ();

	/* register callbacks */
	pan_callback_add (serverlist_get_server_activated_callback(),
	                  grouplist_server_activated_cb, NULL);
	pan_callback_add (group_get_groups_changed_callback(),
	                  grouplist_groups_changed_cb, NULL);
	pan_callback_add (grouplist_group_selection_changed,
	                  grouplist_selection_changed_cb, NULL);
	pan_callback_add (server_get_groups_added_callback(),
	                  grouplist_server_groups_added_cb, NULL);
	pan_callback_add (server_get_groups_removed_callback(),
	                  grouplist_server_groups_removed_cb, NULL);

       	toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL,GTK_TOOLBAR_TEXT);

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GNOME_PAD_SMALL); 
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);

	Pan.group_clist = gtk_clist_new_with_titles (5, titles);
	clist = GTK_CLIST(Pan.group_clist);
	clist->button_actions[1] = clist->button_actions[0];
	pan_widget_set_font (GTK_WIDGET(Pan.group_clist), grouplist_font);
	gtk_clist_set_column_justification (clist, 2, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification (clist, 3, GTK_JUSTIFY_RIGHT);


	w = gtk_label_new (_("Groups"));
	gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar), w, "", "");
	gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

	grouplist_mode_menu = grouplist_mode_menu_create();
	gtk_box_pack_start (GTK_BOX(hbox), grouplist_mode_menu, FALSE, FALSE, 0);

	w = gtk_label_new (_("Find:"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);

	groupname_filter_entry = w = gnome_entry_new ("Grouplist Toolbar Filter");
	gnome_entry_set_max_saved (GNOME_ENTRY(w), 10);
	gtk_signal_connect (GTK_OBJECT(gnome_entry_gtk_entry(GNOME_ENTRY(w))), "activate",
			    GTK_SIGNAL_FUNC(grouplist_filter_changed_cb_nolock), w);
	gtk_signal_connect (GTK_OBJECT(GTK_COMBO(w)->popwin), "button_press_event",
			    GTK_SIGNAL_FUNC(grouplist_filter_changed_cb_nolock), w);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w,
		_("Type in a group search string and press ENTER.  "
		  "Wildcards are allowed."), "");
	gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 100);
	gtk_widget_show_all (toolbar);
	gtk_box_pack_start (GTK_BOX(hbox), w, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	w = gtk_scrolled_window_new (NULL, NULL);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (w),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(w), Pan.group_clist);
        gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "select_row",
			    GTK_SIGNAL_FUNC (select_row_cb), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "unselect_row",
			    GTK_SIGNAL_FUNC (select_row_cb), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "button_press_event",
			    GTK_SIGNAL_FUNC (grouplist_button_press), NULL);
	gtk_signal_connect (GTK_OBJECT (Pan.group_clist), "key_press_event",
	                    GTK_SIGNAL_FUNC (grouplist_key_press),
	                    NULL);


	gtk_clist_set_column_width     (clist, 0, 16);
	gtk_clist_set_column_min_width (clist, 0, 16);
	gtk_clist_set_column_max_width (clist, 0, 16);

	gtk_clist_set_column_width (clist, 1, 170);
	gtk_clist_set_column_width (clist, 2, 36);
	gtk_clist_set_column_width (clist, 3, 33);
	gtk_clist_set_column_width (clist, 4, 350);

	gtk_clist_set_selection_mode (clist, GTK_SELECTION_EXTENDED);

	/* Create the Popup Menu while we are here. */
	group_clist_menu = gnome_popup_menu_new (group_menu);	

	/* Lets have the Pixmap ready too */
	folder_pixmap = (GnomePixmap*) gnome_pixmap_new_from_xpm_d (folder_xpm);
	subscribed_pixmap = GNOME_PIXMAP(gnome_stock_pixmap_widget_new (Pan.window, GNOME_STOCK_MENU_BOOK_RED));

	/* bootstrap */
	grouplist_set_server_nolock (serverlist_get_active_server());

	gtk_widget_show_all (vbox);
	debug_exit ("grouplist_create");
	return vbox;
}

static gchar * _last_filter_str = NULL;

static void
grouplist_filter_changed_cb_nolock (void)
{
	GtkWidget * e = gnome_entry_gtk_entry(GNOME_ENTRY(groupname_filter_entry));
	gchar * pch = gtk_editable_get_chars (GTK_EDITABLE(e), 0, -1);

	if (!pan_strcmp (pch, _last_filter_str))
		g_free (pch);
	else {
		replace_gstr (&_last_filter_str, pch);
		grouplist_refresh_filter_nolock ();
	}
}

/*****
******
*****/

static void
grouplist_add_group_nolock (const Group * group)
{
	gint row;
	GnomePixmap * pixmap;
	gchar total_buf[16] = { '\0' };
	gchar unread_buf[16] = { '\0' };
	gchar name_buf[256] = { '\0' };
	const gint total = group->article_qty;
	const gint unread = MAX(0, total - group->article_read_qty);
	gchar * text[5];

	/* commatize the totals */
	commatize_ulong (total, total_buf);
	commatize_ulong (unread, unread_buf);

	/* look for special qualities of the group */
	grouplist_get_group_readable_name (group, name_buf, sizeof(name_buf));
       
	text[0] = "";
	text[1] = name_buf;
	text[2] = unread_buf;
	text[3] = total_buf;
	text[4] = (gchar*)(group->description?group->description:"");

	/* add the row */
	row = gtk_clist_prepend (GTK_CLIST(Pan.group_clist), (char**)text);
	gtk_clist_set_row_data (GTK_CLIST(Pan.group_clist), row, (gpointer)group);

	/* add the pixmap, if any */
	pixmap = NULL;
	if (group_is_folder (group))
		pixmap = folder_pixmap;
	else if (group_is_subscribed (group))
		pixmap = subscribed_pixmap;
	if (pixmap != NULL)
		gtk_clist_set_pixmap (GTK_CLIST(Pan.group_clist), row, 0,
		                      pixmap->pixmap,
		                      pixmap->mask);
}

static void
grouplist_rebuild_nolock (GPtrArray * groups, GPtrArray * sel)
{
	gint i;
	GtkCList * clist = GTK_CLIST (Pan.group_clist);

	gtk_clist_freeze (clist);
	gtk_clist_clear (clist);

	for (i=groups->len-1; i>=0; --i)
	{
		Group * group = GROUP(g_ptr_array_index (groups, i));

		if (is_nonempty_string(groupname_filter_string) && fnmatch(groupname_filter_string,group->name,PAN_CASEFOLD))
			continue;

		grouplist_add_group_nolock (group);

	}

	if (sel!=NULL && sel->len!=0)
		grouplist_set_selected_groups_nolock (sel);

	gtk_clist_thaw (clist);
}

static gint
grouplist_update_ui_mainthread (gpointer unused)
{
	gchar * title;
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GPtrArray * groups;
	GPtrArray * sel;
	debug_enter ("grouplist_update_ui_mainthread");

	/**
	***  Get the list of groups that we're going to add.
	***
	***  Note that we temporarily turn off the groups_added
	***  callback -- the groups may be loaded as a result of
	***  our call, and we don't want to update the clist twice.
	***  This is a temporary workaround that should dissapear
	***  in 0.10.x
	**/
	pan_callback_remove (server_get_groups_added_callback(),
	                     grouplist_server_groups_added_cb, NULL);
	if (group_mode == GROUP_SUBSCRIBED)
	{
		groups = my_server==NULL
			? g_ptr_array_new ()
			: server_get_groups (my_server, SERVER_GROUPS_SUBSCRIBED);
	}
	else if (group_mode == GROUP_ALL)
	{
		groups = my_server==NULL
			? g_ptr_array_new ()
			: server_get_groups (my_server, SERVER_GROUPS_ALL);
	}
	else if (group_mode == GROUP_FOLDERS)
	{
		Server * s = serverlist_get_named_server(INTERNAL_SERVER_NAME);
		groups = s==NULL
			? g_ptr_array_new()
			: server_get_groups (s, SERVER_GROUPS_ALL);
	}
	else if (group_mode == GROUP_NEW)
	{
		groups = g_ptr_array_new ();
		if (my_server != NULL) {
			guint i;
			GPtrArray * a = server_get_groups (my_server, SERVER_GROUPS_ALL);
			for (i=0; i<a->len; ++i) {
				Group * g = GROUP(g_ptr_array_index(a,i));
				if (group_is_new (g))
					g_ptr_array_add (groups, g);
			}
			g_ptr_array_free (a, TRUE);
		}
	}
	else
	{
		pan_warn_if_reached ();
		groups = g_ptr_array_new ();
	}
	pan_callback_add (server_get_groups_added_callback(),
	                  grouplist_server_groups_added_cb, NULL);

	/* get the list of selected groups */
       	sel = grouplist_get_selected_groups ();

	/* set the groups column title */
	switch (group_mode) {
		case GROUP_SUBSCRIBED: title = _("Subscribed Groups");  break;
		case GROUP_ALL:        title = _("All Groups");         break;
		case GROUP_NEW:        title = _("New Groups");         break;
		case GROUP_FOLDERS:    title = _("Folders");            break;
		default:               title = _("Bug in Code");        break;
	}

	/* update UI */
	pan_lock_unconditional ();
	g_static_mutex_lock (&grouplist_ui_mutex);
	gtk_clist_set_column_title (clist, 1, title);
	grouplist_rebuild_nolock (groups, sel);
	g_static_mutex_unlock (&grouplist_ui_mutex);
	pan_unlock_unconditional ();

	/* cleanup */
	g_ptr_array_free (sel, TRUE);
	g_ptr_array_free (groups, TRUE);
	debug_exit ("grouplist_update_ui_mainthread");
	return 0;
}

void
grouplist_refresh_nolock (void)
{
	Server * server = grouplist_get_visible_server ();

	if (server!=NULL && server==grouplist_get_visible_server())
		run_in_main_thread_nolock (grouplist_update_ui_mainthread, NULL);
}

/* firing this callback is delegated to here because you should never fire
 * a callback inside a pan_lock. */
static void*
fire_server_changed_callback (void * unused)
{
	pan_callback_call (grouplist_server_set_callback, my_server, NULL);
	return NULL;
}

static void
grouplist_set_server_nolock (Server * server)
{
	Server * old_server;
	debug_enter ("grouplist_set_server_nolock");

	/* find out what server we have, and what kind of groups it has */
	old_server = my_server;
	my_server = server;
	run_in_worker_thread (fire_server_changed_callback, NULL);

	if (my_server==NULL)
	{
		if (grouplist_get_visible_server()==old_server)
			gtk_clist_clear (GTK_CLIST(Pan.group_clist));
	}
	else
	{
		gint view_mode;
		GPtrArray * groups;
		const ServerGroupsType have_type = file_grouplist_exists (my_server);

		/* if we have subbed, always load those, otherwise load all */
		if (have_type & SERVER_GROUPS_SUBSCRIBED)
		{
			view_mode = GROUP_SUBSCRIBED;

			groups = my_server==NULL
				? g_ptr_array_new ()
				: server_get_groups (my_server, SERVER_GROUPS_SUBSCRIBED);
		}
		else if (have_type)
		{
			view_mode = GROUP_ALL;

			groups = my_server==NULL
				? g_ptr_array_new ()
				: server_get_groups (my_server, SERVER_GROUPS_ALL);
		}
		else
		{
			gchar * s;
			GtkWidget * w;

			view_mode = GROUP_ALL;
			groups = NULL;

			if (grouplist_get_visible_server()==old_server)
				gtk_clist_clear (GTK_CLIST(Pan.group_clist));

			s = g_strdup_printf (
				_("We don't have a group list for server \"%s\".\nShall we get one?"),
				server->name);
			w = gnome_message_box_new (s, GNOME_MESSAGE_BOX_QUESTION,
						      GNOME_STOCK_BUTTON_YES,
						      GNOME_STOCK_BUTTON_NO, NULL);
			gnome_dialog_set_default (GNOME_DIALOG(w), 0);
			gnome_dialog_set_parent (GNOME_DIALOG(w),
						 GTK_WINDOW(Pan.window));
			gtk_signal_connect (GTK_OBJECT(w), "clicked",
					    GTK_SIGNAL_FUNC(group_dialog_clicked_cb),
					    server);
			gtk_widget_show_all (w);

			g_free (s);
		}

		/* update the UI */
		grouplist_set_view_mode_nolock_impl (view_mode);
		if (groups != NULL)
			grouplist_rebuild_nolock (groups, NULL);

		/* cleanup */
		if (groups != NULL)
			g_ptr_array_free (groups, TRUE);
	}

	debug_exit ("grouplist_set_server_nolock");
}


/***
****
****
***/

/**
 * @group: Group who's row is to be updated.
 *
 * Update the CList row for a group to show it's message counts or
 * any other relative data (subscribed/on disk) for that group.
 **/
static void
grouplist_update_group_nolock (const Group * group)
{
	GtkCList * list;
	gint row;

	/* sanity check */
	g_return_if_fail (group != NULL);

	list = GTK_CLIST (Pan.group_clist);
       	row = gtk_clist_find_row_from_data (list,(gpointer)group);
	if (row != -1)
	{
		gchar buf[256] = { '\0' };
		const gint total = group->article_qty;
		const gint unread = MAX(0, total-group->article_read_qty);

		/* icon column */
		if (group_is_subscribed (group))
			gtk_clist_set_pixmap (GTK_CLIST(Pan.group_clist), row, 0,
			                      subscribed_pixmap->pixmap,
			                      subscribed_pixmap->mask);
		else if (!group_is_folder(group))
			gtk_clist_set_text (GTK_CLIST(Pan.group_clist),
					    row, 0, "");

		/* name column */
		grouplist_get_group_readable_name (group, buf, sizeof(buf));
		gtk_clist_set_text (list, row, 1, buf);

		/* total column */
		commatize_ulong (total, buf);
		gtk_clist_set_text (list, row, 3, buf);

		/* unread column */
		commatize_ulong (unread, buf);
		gtk_clist_set_text (list, row, 2, buf);
	}
}

static gint
grouplist_update_groups_mainthread (gpointer data)
{
	guint i;
	GPtrArray * a = (GPtrArray*)data;
	GtkCList * list = GTK_CLIST (Pan.group_clist);
	debug_enter ("grouplist_update_groups_mainthread");

	g_static_mutex_lock (&grouplist_ui_mutex);
	pan_lock_unconditional ();
	gtk_clist_freeze (list);
	for (i=0; i!=a->len; ++i)
		grouplist_update_group_nolock (GROUP(g_ptr_array_index(a,i)));
	gtk_clist_thaw (list);
	pan_unlock_unconditional ();
	g_static_mutex_unlock (&grouplist_ui_mutex);

	/* cleanup */
	g_ptr_array_free (a, TRUE);
	debug_exit ("grouplist_update_groups_mainthread");
	return 0;
}

static void
grouplist_update_groups (const Group ** groups, int group_qty)
{
	gint i;
	GPtrArray * a;
	Server * server;
	debug_enter ("grouplist_update_groups");

	g_return_if_fail (groups != NULL);
	g_return_if_fail (group_qty >= 1);

	/* build an array of groups that we might need to update */
	a = g_ptr_array_new ();
	server = grouplist_get_visible_server ();
	for (i=0; i<group_qty; ++i)
		if (groups[i]!=NULL && groups[i]->server==server)
			g_ptr_array_add (a, (gpointer)groups[i]);

	/* if we have any that we should update, queue it */
	if (a->len != 0) {
		pan_lock ();
		run_in_main_thread_nolock (grouplist_update_groups_mainthread, a);
		pan_unlock ();
	} else {
		g_ptr_array_free (a, TRUE);
	}

	debug_exit ("grouplist_update_groups");
}

/***
****
****
***/

static void
grouplist_internal_select_row (gint      row,
                               gboolean  activate,
                               gboolean  lock)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);

	if (activate)
		button_click_count = 2;

	if (lock) pan_lock ();
	gtk_clist_freeze (list);
	gtk_clist_unselect_all (list);
	gtk_clist_select_row (list,row,-1);
	gtk_clist_thaw (list);
	if (lock) pan_unlock ();
}

static gboolean
grouplist_select_row_if_unread (int row,
                                gboolean activate,
                                gboolean lock)
{
        GtkCList * list = GTK_CLIST(Pan.group_clist);
	Group * g = gtk_clist_get_row_data(list,row);
	const gulong total = g->article_qty;
	const gulong read = g->article_read_qty;
	gboolean retval = FALSE;

	if (read < total) {
		grouplist_internal_select_row (row, activate, lock);
		retval = TRUE;
	}

	return retval;
}
static void
grouplist_next_unread_group (gboolean activate)
{
        GtkCList* list;
	int row_qty, i, row=0;

	pan_lock ();
	list = GTK_CLIST(Pan.group_clist);
        row_qty = list->rows;

	if (row_qty != 0)
	{
		/* get the first row to check */
		row = 0;
		if (list->selection) {
			int sel = GPOINTER_TO_INT(list->selection->data);
			row = (sel + 1) % row_qty;
		}

		/* walk through */
		for ( i=0; i<row_qty; ++i, row=(row+1)%row_qty)
			if (grouplist_select_row_if_unread (row, activate, 0))
				break;
	}

	pan_unlock ();
}
void
grouplist_select_next_unread_group (void)
{
	grouplist_next_unread_group (FALSE);
}
void
grouplist_activate_next_unread_group (void)
{
	grouplist_next_unread_group (TRUE);
}
static void
grouplist_next_group (gboolean activate)
{
        GtkCList* list;
	int row_qty;
	int row = 0;

	pan_lock ();
	list = GTK_CLIST(Pan.group_clist);
        row_qty = list->rows;

	if (row_qty != 0)
	{
		/* get the first row to check */
		row = 0;
		if (list->selection) {
			int sel = GPOINTER_TO_INT(list->selection->data);
			row = (sel + 1) % row_qty;
		}

		grouplist_internal_select_row (row, activate, 0);
	}

	pan_unlock ();
}
void
grouplist_select_next_group (void)
{
	grouplist_next_group (FALSE);
}
void
grouplist_activate_next_group (void)
{
	grouplist_next_group (TRUE);
}

/***
****
***/

static void
group_menu_popup (GdkEventButton *event)
{
	debug_enter ("group_menu_popup");

	/* We must be dealing with real groups */
	{
		pan_lock ();

		/* show the menu */
		gtk_menu_popup (GTK_MENU(group_clist_menu),
				NULL, NULL, NULL, NULL,
				event->button, event->time);
		pan_unlock ();
	} 

	debug_exit ("group_menu_popup");
}

static void
grouplist_refresh_filter_nolock (void)
{
	gchar * new_filter;
        gchar * pch;

	pch  = gtk_editable_get_chars (GTK_EDITABLE(gnome_entry_gtk_entry(GNOME_ENTRY(groupname_filter_entry))), 0, -1);

	/* build the new filter string */
	new_filter = NULL;
	if (is_nonempty_string (pch)) {
		if (strchr(pch,'*') != NULL)
			new_filter = g_strdup (pch);
		else 
			new_filter = g_strdup_printf("*%s*",pch);
	}

	/* if it differs from the old filter string, refresh */
	if (pan_strcmp (new_filter, groupname_filter_string)) {
		replace_gstr (&groupname_filter_string, g_strdup(new_filter));
		grouplist_refresh_nolock ();
	}

	/* cleanup */
	g_free (pch);
	g_free (new_filter);
}
