/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000, 2001  Pan Development Team <pan@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>

#include <gnome.h>

#include <pan/base/article.h>
#include <pan/base/pan-glib-extensions.h>

#include <pan/globals.h>
#include <pan/gui.h>
#include <pan/prefs.h>
#include <pan/print.h>
#include <pan/text.h>
#include <pan/util.h>


/* private structure for a print dialog */
typedef struct
{
	GtkWidget  * window;           /* Gnomedialog */
	GtkWidget  * print_command;    /* entry box, should likely be 'lpr' */
	Article    * article;
}
PrintDialog;


static void print_dialog_print_clicked (PrintDialog *dialog);
static void print_dialog_destroy (PrintDialog *dialog);
static gint print_execute (const gchar *command, const gchar *message);
static void print_dialog_cb (GtkWidget *widget, gint num, gpointer data);
static void print_dialog (Article * article);


/* Thanks to Tim Roberts for the rewrite of this function.
 * @return 0 for success, -1 for any error encountered
 */
static gint
print_execute (const gchar *command, const gchar *message)
{
        FILE * fp = popen (command, "w");
	if (!fp) {
		g_warning ("print_execute(): error in popen()\n");
		return -1;
	}

	/* dump the message to the pipe */
	fputs (message, fp);
	fflush (fp);
	fclose (fp);
	return 0;
}


/* Destroy the dialog and free used memory */
static void
print_dialog_destroy (PrintDialog *dialog)
{
	gtk_widget_destroy (dialog->window);

	group_unref_articles (dialog->article->group, NULL);
	g_free (dialog);
}


/* The user pressed the Print button, build the message and call for it
 * to be printed with the provided print command
 */
static void
print_dialog_print_clicked (PrintDialog * dialog)
{
	gchar * text;
	gchar * command;
	const Article * article;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (dialog->article != NULL);

	/* get information for what we'll be printing */
       	article = dialog->article;

	/* build the message we'll be printing */
	text = article_get_message (article);

	/* find out how to execut it */
	command = gtk_editable_get_chars (GTK_EDITABLE(dialog->print_command),
	                                  0,
	                                  -1);

	/* execute */
	if (print_execute (command, text))
		pan_error_dialog_parented (dialog->window,
			"Printing of message \"%s\" failed!",
			article_get_subject(article));
	else
		gnome_config_set_string ("/Pan/Print/print_command", command);

	/* clean up */
	print_dialog_destroy (dialog);
	g_free (text);
	g_free (command);
}


/* The user pressed a button. Lets like, do things and stuff. Dude. */
static void
print_dialog_cb (GtkWidget * widget_notused,
                 gint         num,
		 gpointer     user_data)
{
	g_return_if_fail (user_data);

	switch (num)
	{
		case 0: /* Print */
			print_dialog_print_clicked (user_data);
			break;
		case 1: /* Cancel */
			print_dialog_destroy (user_data);
			break;
		default:
			pan_warn_if_reached ();
			break;
	}
}


/* Make the dialog with Print and Cancel buttons */
static void
print_dialog (Article * article)
{
	gchar * p;
	GtkWidget * table;
	GtkWidget * label;
	PrintDialog * dialog;

	g_return_if_fail (article != NULL);

	/* make sure the Article stays alive during this dialog */
	group_ref_articles (article->group, NULL);

	/* dialog object */
	dialog = g_new0 (PrintDialog, 1);
	dialog->article = article;
	dialog->window = gnome_dialog_new (_("Pan: Print"),
					   _("Print"),
					   GNOME_STOCK_BUTTON_CANCEL, NULL);

	/* dialog table */
	table = gtk_table_new (1, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(table), GNOME_PAD_SMALL);
        gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
        gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);

	/* print command label */
	label = gtk_label_new (_("Print Command:"));
	gtk_misc_set_alignment (GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table),
	                  label,
	                  0, 1, 0, 1,
	                  GTK_FILL, GTK_FILL,
	                  4, 0);

	/* print command edit field */
	dialog->print_command = gtk_entry_new ();
	p = gnome_config_get_string ("/Pan/Print/print_command");
	if (p == NULL)
		p = g_strdup ("lpr");
	gtk_entry_set_text (GTK_ENTRY(dialog->print_command), p);
	g_free (p);
	gtk_table_attach (GTK_TABLE(table),
	                  dialog->print_command,
	                  1, 2, 0, 1,
	                  GTK_FILL, GTK_FILL,
	                  4, 0);

	/* add the main dialog table to the dialog */
	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog->window)->vbox),
	                   table);
	gtk_signal_connect (GTK_OBJECT(dialog->window),
	                    "clicked",
	                    GTK_SIGNAL_FUNC(print_dialog_cb),
	                    dialog);

	/* show the dialog */
	gnome_dialog_set_parent (GNOME_DIALOG(dialog->window),
	                         GTK_WINDOW(Pan.window));
	gtk_widget_show_all (dialog->window);
}


/**
 * print_cb:
 * @data: the Article* to be printed
 *
 * The only public routine, gets called from the Print button on the toolbar
 * of a message window or File->Print from any menu in Pan.
 **/
void
print_cb (GtkWidget * widget_notused,
          gpointer    data)
{
	Article * a = get_current_article ();

	if (data != NULL)
	{
		print_dialog (ARTICLE(data));
	}
	/* if they aren't printing from a message window,
	   use the current article in the message pane/page */
	else if (!data && a)
	{
		print_dialog (a);
	}
	else if (!data && !a)
	{
		pan_error_dialog (_("You haven't loaded an article to print."));
		return;
	}
}

#ifdef HAVE_GNOME_PRINT

void
pan_gnome_print_cb (void)
{
	/* FIXME(csk): gnome-print stuff goes here */
	print_cb (NULL, NULL);
}

void
pan_gnome_print_preview_cb (void)
{
	/* FIXME(csk): gnome-print stuff goes here */
	print_cb (NULL, NULL);
}

#endif
